/*
    File:       nmea.cpp
    Version:    0.2.0
    Date:       Feb. 23, 2013 - Jul. 07, 2020
	License:	GPL v2
    
	NMEA GPS content parser
    
    ****************************************************************************
    Copyright (C) 2013 Radu Motisan  <radu.motisan@gmail.com>
	
	http://www.pocketmagic.net

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ****************************************************************************
 */

#include "drv_nmea.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>


/*
 * returns base-16 value of chars '0'-'9' and 'A'-'F';
 * does not trap invalid chars!
 */	
uint8_t NMEA::digit2dec(char digit) {
	if (int(digit) >= 65) 
		return int(digit) - 55;
	else 
		return int(digit) - 48;
}

/* returns base-10 value of zero-terminated string
 * that contains only chars '+','-','0'-'9','.';
 * does not trap invalid strings! 
 */
double NMEA::string2double(char* s) {
	long  integer_part = 0;
	double decimal_part = 0.0;
	double decimal_pivot = 0.1;
	bool isdecimal = false, isnegative = false;
	
	char c;
	while ( ( c = *s++) )  { 
		// skip special/sign chars
		if (c == '-') { isnegative = true; continue; }
		if (c == '+') continue;
		if (c == '.') { isdecimal = true; continue; }
		
		if (!isdecimal) {
			integer_part = (10 * integer_part) + (c - 48);
		}
		else {
			decimal_part += decimal_pivot * (double)(c - 48);
			decimal_pivot /= 10.0;
		}
	}
	// add integer part
	decimal_part += (double)integer_part;
	
	// check negative
	if (isnegative)  decimal_part = - decimal_part;

	return decimal_part;
}
uint8_t NMEA::hexdigit(uint8_t digit) {
	if (digit >= '0' && digit <= '9') return digit  - '0';
    else if (digit >= 'a' && digit <='f') return digit - ('a' - 10);
    else if (digit >= 'A' && digit <='F') return digit - ('A' - 10);
    else return 0; //error
}

// take a hex string and convert it to a 32bit number (max 8 hex digits)
uint32_t NMEA::hex2int(char *hex) {
    uint32_t val = 0;
     while (*hex) {
    	// shift 4 to make space for new digit, and add the 4 bits of the new digit
    	val = (val << 4) | (hexdigit(*hex++) & 0xF);
    }
    return val;
}

double NMEA::decodeCoordinate(double value) {
	double degrees = trunc(value / 100.0f);
	double minutes = value - (degrees * 100.0f);
	return degrees + minutes / 60.0f;
}


bool NMEA::parsedata(uint8_t messageType) {
	ready = false;
	// used to cut words
	uint8_t start = 0,
			stop = 0,
			wi=0;

	for (int i=0;i<bufIndex;i++) {
		// compute simple checksum on all nmea bytes, starting AFTER $ and ending BEFORE *, both exclusive
		if (i > 0 && i < starIndex)  checksum ^= buf[i];

		// mark separator: stop word is either , * or end of line \r
		if (buf[i] == ',' || buf[i] == '*' || i == bufIndex-1) {
			stop = i;
			mstrncpy(word, (char *)buf + start +1, MIN( (stop - start - (buf[i] == ',')), MAX_WORD )); // cut word
			start = stop;

			// word is ready, interpret it
			if (messageType == 1) switch (wi) {
				case 0: break;
				case 1:
					hour = digit2dec(word[0]) * 10 + digit2dec(word[1]);
					min = digit2dec(word[2]) * 10 + digit2dec(word[3]);
					sec = digit2dec(word[4]) * 10 + digit2dec(word[5]);
				break;
				case 2:
					latitude = string2double(word);
					break;
				case 3:
					if (word[0] == 'S') latitude  *= -1.0;
					break;
				case 4:
					longitude = string2double(word);
					break;
				case 5:
					if (word[0] == 'W') longitude *= -1.0;
					break;
				case 6: //  GPS quality indicator (0=invalid; 1=GPS fix; 2=Diff. GPS fix)
					fix = word[0] - '0';
					if (fix == 0) return false;
				break;
				case 7: //Number of satellites in use [not those in view]
					sats = (word[0] - '0') * 10 + word[1] - '0';
					break;
				case 9:
					altitude = string2double(word);
					break;
				case 15:
						if (hex2int(word) != checksum) return false; // Checksum failed
					break;
			}
			else if (messageType == 2) switch (wi) {
				case 0: break;
				case 1:
					hour = digit2dec(word[0]) * 10 + digit2dec(word[1]);
					min = digit2dec(word[2]) * 10 + digit2dec(word[3]);
					sec = digit2dec(word[4]) * 10 + digit2dec(word[5]);
				break;
				case 2:
					if (word[2] == 'V')  return false;
				break;
				case 3:
					latitude = string2double(word);
					break;
				case 4:
					if (word[0] == 'S') latitude  *= -1.0;
					break;
				case 5:
					longitude = string2double(word);
					break;
				case 6:
					if (word[0] == 'W') longitude *= -1.0;
					break;
				case 7:
					speed = string2double(word) * 1.852; // convert knots to km/h
				break;
				case 9:
					day = digit2dec(word[0]) * 10 + digit2dec(word[1]);
					month = digit2dec(word[2]) * 10 + digit2dec(word[3]);
					year = digit2dec(word[4]) * 10 + digit2dec(word[5]);
				break;
			}
			wi++; // word index to know the function
		}
	}
	// convert latitude and longitude
	latitude = decodeCoordinate(latitude);
	longitude = decodeCoordinate(longitude);

	// ready to use
	ready = true;
}


// On a microcontroller, you would send data from the UART Interrupt
uint8_t NMEA::fusedata(char c) {
	buf[bufIndex] = c;

 	if (bufIndex == 0 && c != '$') {
 		//ready = false;
		bufIndex = 0; return 0;
	}
	if (c == '*') starIndex = bufIndex;

	if (c== '\r') {
		if (mstrncmp((char*)buf, "$GPGGA", 6) == 0) parsedata(1);
		if (mstrncmp((char*)buf, "$GPRMC", 6) == 0) parsedata(2);
		bufIndex = 0;
		return 0;
	}

	bufIndex ++;

	if (bufIndex == MAX_BUFFER - 1) {
		//ready = false;
		bufIndex = 0; // start over
	}

	return bufIndex;
}

// string functions

uint8_t NMEA::mstrncmp( char* s1,  char* s2, uint8_t n) {
    while(n--)
        if(*s1++!=*s2++)
            return *(unsigned char*)(s1 - 1) - *(unsigned char*)(s2 - 1);
    return 0;
}

char* NMEA::mstrncpy(char *dest,  char *src, uint8_t n) {
    char *ret = dest;
    do {
        if (!n--) return ret;
    } while (*dest++ = *src++);
    while (n--) *dest++ = 0;
    return ret;
}

		
bool NMEA::isdataready() {
	return ready;
}

uint8_t NMEA::getHour() {
	return hour;
}	
uint8_t NMEA::getMinute() {
	return min;
}
uint8_t NMEA::getSecond() {
	return sec;
}
uint8_t NMEA::getDay() {
	return day;
}
uint8_t NMEA::getMonth() {
	return month;
}
uint8_t NMEA::getYear() {
	return year;
}

double NMEA::getLatitude() {
	return ready?latitude:0;
}

double NMEA::getLongitude() {
	return ready?longitude:0;
}

uint8_t NMEA::getSatellites() {
	return ready?sats:0;
}

double  NMEA::getAltitude() {
	return ready?altitude:0;
}

double NMEA::getSpeed() {
	return ready?speed:0;
}
